module.exports = {
    DEFAULT_TRANSFER_AMOUNT: "1000",
    DEFAULT_ACCOUNT_FUND_AMOUNT: "300000000",
    MAX_NFTOKEN_PAGE_OBJECTS_LIMIT: 32,
    DEFAULT_TEST_GENESIS_ACCOUNT_FUND: "100000000000",
    DEFAULT_DELETE_ACCOUNT_FEE: "2000000",
    BASE_RESERVE: "10000000",
    OWNER_RESERVE: "2000000",
};
