module.exports = {
    connectionScheme: '',
    connectionUrl: '',
    connectionType: '',
    client: '',
    crypto_algorithm: '',
    sourceWallet: '',
    sourceAddress: '',
    sourceBalance: '',
    destinationWallet: '',
    destinationAddress: '',
    destinationBalance: '',
    issuerWallet: '',
    issuerAddress: '',
    issuerBalance: '',
    aliceWallet: '',
    aliceAddress: '',
    aliceBalance: '',
    bobWallet: '',
    bobAddress: '',
    bobBalance: '',
    carolWallet: '',
    carolAddress: '',
    carolBalance: '',
    invalidWallet: '',
    invalidAddress: '',
    invalidBalance: '',
    minDrop: '',
    transferAmount: '',
    payment: '',
    prepared: '',
    signed_payment_tx: '',
    payment_response: '',
    response: '',
    account_set_response_1: '',
    account_set_response_2: '',
    py_response_1: '',
    py_response_2: '',
    py_response_3: '',
    tc_response_1: '',
    ts_response_1: '',
    ts_response_2: '',
    ts_response_3: '',
    responses: '',
    offer_create_response_1: '',
    offer_create_response_2: '',
    nft_mint_response_1: '',
    nft_mint_response_2: '',
    nft_burn_response_1: '',
    nft_burn_response_2: '',
    exception: '',
    testStatus: ''
};
